import time
from enum import Enum
from threading import Thread
from typing import List, Callable, Optional, Any, Tuple, Union

from py_cui import PyCUI
from py_cui import popups
from py_cui.widget_set import WidgetSet

from qrogue.game.logic.actors import Boss, Enemy, Riddle, Challenge, Robot, FusionBot
from qrogue.game.logic.actors.puzzles import FusionTarget
from qrogue.game.logic.collectibles import Instruction
from qrogue.game.world.dungeon_generator.wave_function_collapse import WFCManager
from qrogue.game.world.map import CallbackPack, Map
from qrogue.game.world.navigation import Direction
from qrogue.game.world.tiles import WalkTriggerTile, Message, Collectible
from qrogue.graphics import WidgetWrapper
from qrogue.graphics.popups import Popup, MultilinePopup, ConfirmationPopup
from qrogue.graphics.rendering import MultiColorRenderer
from qrogue.graphics.widgets import Renderable, BossFightWidgetSet, ExploreWidgetSet, \
    FightWidgetSet, MenuWidgetSet, MyWidgetSet, NavigationWidgetSet, PauseMenuWidgetSet, RiddleWidgetSet, \
    ChallengeWidgetSet, WorkbenchWidgetSet, TrainingsWidgetSet, Widget, TransitionWidgetSet, \
    ScreenCheckWidgetSet, LevelSelectWidgetSet, FusionCircuitWidgetSet
from qrogue.util import common_messages, CheatConfig, Config, GameplayConfig, UIConfig, HelpText, \
    Logger, PathConfig, Controls, Keys, RandomManager, PyCuiConfig, PopupConfig, PyCuiColors, Options, \
    CommonInfos, MapConfig, WordingConfig
from qrogue.util.achievements import Achievement, Unlocks
from qrogue.util.game_simulator import GameSimulator
from qrogue.util.key_logger import KeyLogger, OverWorldKeyLogger, DummyKeyLogger
from .map_management import MapManager
from .save_data import NewSaveData


class QrogueCUI(PyCUI):
    # how many seconds to wait after starting to mark initialization as complete (relevant for input handling)
    __INIT_DELAY = 0.1

    class _State(Enum):
        Menu = 0
        Pause = 1
        Explore = 2
        Fight = 3
        # Shop = 4
        Riddle = 5
        BossFight = 6

        # Spaceship = 7
        Workbench = 8
        # Navigation = 9
        Training = 10

        Challenge = 11

        Transition = 12  # for atmospheric transitions and elements
        ScreenCheck = 13  # a menu to check dimensions and color of the screen/terminal the game is played
        LevelSelect = 14

        FusionCircuit = 25

    class _StateMachine:
        def __init__(self, renderer: "QrogueCUI"):
            self.__renderer = renderer
            self.__cur_state: Optional["QrogueCUI._State"] = None
            self.__prev_state: Optional["QrogueCUI._State"] = None

        @property
        def cur_state(self) -> "QrogueCUI._State":
            return self.__cur_state

        @property
        def prev_state(self) -> "QrogueCUI._State":
            return self.__prev_state

        def change_state(self, state: "QrogueCUI._State", data) -> None:
            if self.__cur_state is not QrogueCUI._State.Transition:
                # don't overwrite previous prev_state if the new one would be Transition
                self.__prev_state = self.__cur_state
            self.__cur_state = state

            if self.__cur_state == QrogueCUI._State.Menu:
                self.__renderer._switch_to_menu(data)
            elif self.__cur_state == QrogueCUI._State.Explore:
                self.__renderer._switch_to_explore(data)
            elif self.__cur_state == QrogueCUI._State.Fight:
                self.__renderer._switch_to_fight(data)
            elif self.__cur_state == QrogueCUI._State.Riddle:
                self.__renderer._switch_to_riddle(data)
            elif self.__cur_state == QrogueCUI._State.Challenge:
                self.__renderer._switch_to_challenge(data)
            elif self.__cur_state == QrogueCUI._State.BossFight:
                self.__renderer._switch_to_boss_fight(data)
            elif self.__cur_state == QrogueCUI._State.Pause:
                self.__renderer._switch_to_pause()

            elif self.__cur_state == QrogueCUI._State.Training:
                self.__renderer._switch_to_training(data)
            elif self.__cur_state == QrogueCUI._State.Workbench:
                self.__renderer._switch_to_workbench(data)
            elif self.__cur_state == QrogueCUI._State.FusionCircuit:
                self.__renderer._switch_to_fusion_circuit(data)

            elif self.__cur_state == QrogueCUI._State.Transition:
                self.__renderer._switch_to_transition(data)
            elif self.__cur_state == QrogueCUI._State.ScreenCheck:
                self.__renderer._switch_to_screen_check(data)

            elif self.__cur_state == QrogueCUI._State.LevelSelect:
                self.__renderer._switch_to_level_select(data)

            else:
                Popup.error("Illegal Game State! It's still possible that the game works fine, so if nothing seems "
                            "wrong you can continue to play. However, we recommend to save the game and restart it. ",
                            add_report_note=True)

    class _PopupHistory:
        def __init__(self, show_popup: Callable[[MultilinePopup], None]):
            self.__show_popup = show_popup
            self.__history: List[MultilinePopup] = []
            self.__index = -1
            self.__remove_on_close = False

        @property
        def is_empty(self) -> bool:
            return len(self) <= 0

        @property
        def present_index(self) -> int:
            return len(self.__history) - 1

        @property
        def is_in_present(self) -> bool:
            return self.__index == self.present_index

        def back(self, show_popup: bool = True):
            if self.is_empty or self.__index == 0:
                return  # we already show the oldest popup (or cannot show anything)
            elif self.__index < 0:
                self.__index = 0  # show oldest popup for invalid indices
            else:
                self.__index -= 1

            if show_popup: self.show()

        def forth(self, show_popup: bool = True):
            if self.is_empty or self.__index == self.present_index:
                return  # we already show the most recent popup (or cannot show anything)
            elif self.__index < 0:
                self.__index = self.present_index  # show most recent popup for invalid indices
            else:
                self.__index += 1

            if show_popup: self.show()

        def jump_to_present(self, show_popup: bool = True):
            if self.is_empty or self.__index == self.present_index:
                return  # we already show the most recent popup (or cannot show anything)
            self.__index = self.present_index

            if show_popup: self.show()

        def show(self):
            if 0 <= self.__index < len(self.__history):
                self.__show_popup(self.__history[self.__index])
            # else: can happen if a level doesn't start with a message, e.g. expeditions

        def add(self, new_popup: MultilinePopup, is_permanent: bool):
            """
            :param new_popup: the popup to add to the history
            :param is_permanent:  permanent popups stay in the history until it is reset, while non-permanent (i.e.,
                temporary) popups will be removed after the popup display closes
            """
            self.__history.append(new_popup)
            self.__index = self.present_index
            self.__remove_on_close = not is_permanent

        def resolve(self, force_remove: bool = False):
            if len(self.__history) <= 0:
                return  # can happen if first Popup of a history is not permanent, hence there is nothing to resolve
            if self.__remove_on_close or force_remove:
                self.__history.pop()
                self.__index = min(self.__index, self.present_index)  # adapt index if it pointed to the removed popup
                self.__remove_on_close = False
            elif 0 <= self.present_index < len(self.__history):  # if is needed to not crash on level loading error
                self.__history[self.present_index].freeze()

        def reset(self):
            self.__history.clear()
            self.__index = -1

        def __len__(self) -> int:
            return len(self.__history)

    @staticmethod
    def start_simulation(simulation_path: str, in_keylog_folder: bool = True,
                         automation_step_time: Optional[int] = None, auto_scroll_transitions: bool = False,
                         stop_when_finished: bool = False) -> Optional[NewSaveData]:
        """

        Args:
            simulation_path: path to the .qrkl file we want to simulate
            in_keylog_folder: whether the given simulation path is inside the user data keylog folder or an absolute
                path
            automation_step_time: None if simulation should be manual, else the number of milliseconds to wait before
                performing the next step
            auto_scroll_transitions: whether level transitions should also continue automatically or not
            stop_when_finished: whether we want to stop the CUI when the simulation is finished or not

        Returns:
            the save data of the simulation or None if the simulation path was invalid (e.g., non-existent or no .qrkl
            file)
        """
        Logger.instance().assertion(automation_step_time is None or automation_step_time > 0,
                                    f"Invalid automation_step_time: {automation_step_time}!")
        try:
            simulator = GameSimulator(simulation_path, in_keylog_folder)
            qrogue_cui = QrogueCUI(simulator.seed, save_data=NewSaveData(simulator.save_state))
            qrogue_cui._set_simulator(simulator, auto_scroll_transitions, stop_when_finished)

            if automation_step_time is not None:
                qrogue_cui.set_refresh_timeout(automation_step_time)

            if simulator.simulates_over_world:
                return qrogue_cui.start()
            else:
                return qrogue_cui.start(simulator.map_name)

        except FileNotFoundError as fnf:
            error_text = f"Simulation file \"{simulation_path}\" not found: {fnf}"
            # only create a popup for the error during manual simulations
            if automation_step_time is None and Popup.is_initialized():
                Popup.error(error_text)
            else:
                Logger.instance().error(error_text, show=False, from_pycui=False)
            return None

    def __init__(self, seed: int, width: int = UIConfig.WINDOW_WIDTH, height: int = UIConfig.WINDOW_HEIGHT,
                 save_data: Optional[NewSaveData] = None):
        self.__init_complete = False
        super().__init__(width, height)
        PyCuiConfig.set_get_dimensions_callback(self._get_absolute_grid_dimensions)
        self.set_title(f"QRogue {Config.version()}")
        self.__controls = Controls(self._handle_key_presses)
        self.__rm = RandomManager.create_new(seed)
        self.__wfc_manager = WFCManager()
        self.__wfc_manager.load()

        def move_focus(_widget: WidgetWrapper, _widget_set: WidgetSet):
            # this check is necessary for manual widget-set switches due to the call-order (the callback happens before
            # this move_focus here)
            if _widget_set is self.__cur_widget_set:
                self.move_focus(_widget, auto_press_buttons=False)

        self._auto_focus_buttons = False
        Widget.set_move_focus_callback(move_focus)

        # INIT MANAGEMENT
        Logger.instance().set_popup(lambda text: Popup.error(text, log_error=False, add_report_note=True))
        CheatConfig.init(self.__show_cheat_info_popup, self.__show_input_popup,
                         deactivate_cheats=not Config.debugging(),
                         allow_cheats=Config.debugging())
        Popup.update_popup_functions(self.__show_message_popup)
        ConfirmationPopup.update_popup_function(self.__show_confirmation_popup)

        self.__cbp = CallbackPack(self.__start_fight, self.__start_boss_fight, self.__open_riddle,
                                  self.__open_challenge, self.__game_over)
        ########################################
        self.__stored_save: Optional[NewSaveData] = None  # temporarily stores the "real" save data when simulating
        self.__save_data = NewSaveData() if save_data is None else save_data

        def start_level_transition(prev_map_name: str, next_map_name: str, callback: Callable[[], None]):
            texts = [
                TransitionWidgetSet.TextScroll.fast(f"Loading {next_map_name}"),
                TransitionWidgetSet.TextScroll.medium(f"..."),
            ]
            auto_scroll = self.__auto_scroll_simulation_transitions
            self.__state_machine.change_state(QrogueCUI._State.Transition, data=(texts, callback, auto_scroll))

        self.__map_manager = MapManager(self.__wfc_manager, self.__save_data, self.__rm.seed, self.__start_level,
                                        start_level_transition, lambda: self._switch_to_menu(None), self.__cbp)
        ########################################

        self.__map_manager.fill_expedition_queue(lambda: None, no_thread=True)

        Popup.update_check_achievement_function(self.__map_manager.check_level_event)
        common_messages.set_show_callback(lambda text: Popup.generic_info(Config.system_name(), text))
        common_messages.set_show_info_callback(Popup.generic_info)
        common_messages.set_ask_callback(ConfirmationPopup.ask)
        WalkTriggerTile.set_show_explanation_callback(Popup.from_message)
        Message.set_show_callback(Popup.from_message_trigger)
        Collectible.set_pickup_message_callback(Popup.generic_info, self.__save_data.check_unlocks)

        self.__ow_key_logger = OverWorldKeyLogger()
        self.__key_logger = KeyLogger()
        self.__simulator: Optional[GameSimulator] = None
        self.__stop_with_simulation_end = False
        self.__auto_scroll_simulation_transitions = False
        self.__last_input = time.time()
        self.__last_key: Optional[int] = None
        self.__focused_widget: Optional[Widget] = self.get_selected_widget()

        # INIT POPUP HISTORY
        def _show_popup_for_history(historic_popup: MultilinePopup):
            self.__focused_widget = self.get_selected_widget()
            self._popup = historic_popup

        self.__popup_history = self._PopupHistory(_show_popup_for_history)

        # INIT WIDGET SETS
        self.__menu = MenuWidgetSet(self.__controls, self.__render, Logger.instance(), self,
                                    self.__map_manager.load_first_uncleared_map, self.__start_playing,
                                    self.__start_expedition, self.stop, self.__show_screen_check,
                                    self.__show_level_select, self.__show_workbench, self.__save_data.check_unlocks,
                                    self._conditional_saving)
        self.__level_select = LevelSelectWidgetSet(
            self.__controls, Logger.instance(), self, self.__render, self.__rm, self.__show_input_popup,
            self.__save_data.get_completed_levels,
            self._switch_to_menu,
            lambda map_name, map_seed, gate_list: self.__map_manager.load_map(map_name, None, map_seed, gate_list),
            lambda: int(self.__save_data.get_progress(Achievement.CompletedExpedition)[0]), self.__save_data.get_gates,
            self.__save_data.check_unlocks
        )
        self.__screen_check = ScreenCheckWidgetSet(self.__controls, Logger.instance(), self, self.__render,
                                                   self._switch_to_menu)
        self.__transition = TransitionWidgetSet(self.__controls, Logger.instance(), self, self.__render,
                                                self.set_refresh_timeout)
        self.__pause = PauseMenuWidgetSet(self.__controls, self.__render, Logger.instance(), self,
                                          self.__general_continue, self._conditional_saving, self._switch_to_menu,
                                          self.__map_manager.reload, self.__save_data.to_achievements_string,
                                          self.__save_data.check_unlocks)
        self.__pause.set_data(None, "Qrogue", None)

        self.__training = TrainingsWidgetSet(self.__controls, self.__render, Logger.instance(), self,
                                             lambda b: None, self.__popup_history.show,
                                             self.__save_data.check_unlocks)  # todo: update signature
        self.__workbench = WorkbenchWidgetSet(Logger.instance(), self, self.__render, self.__controls,
                                               lambda: self.__save_data.num_quantum_fusers,
                                              self.__save_data.get_original_gates, self.__save_data.add_gate,
                                              self.__save_data.decompose, self.__save_data.check_unlocks,
                                              self._switch_to_menu, self.__show_fusion_circuit)
        self.__fusion_circuit = FusionCircuitWidgetSet(self.__controls, self.__render, Logger.instance(), self,
                                                       self.__continue_explore, self.__popup_history.show,
                                                       self.__save_data.check_unlocks,
                                                       lambda: self.__save_data.num_quantum_fusers,
                                                       self.__save_data.add_gate,
                                                       self.__save_data.decompose, self.__show_input_popup,
                                                       self.__show_workbench)

        self.__navigation = NavigationWidgetSet(self.__controls, self.__render, Logger.instance(), self)

        self.__explore = ExploreWidgetSet(self.__controls, self.__render, Logger.instance(), self)
        self.__fight = FightWidgetSet(self.__controls, self.__render, Logger.instance(), self, self.__continue_explore,
                                      self.__popup_history.show, self.__save_data.check_unlocks)
        self.__boss_fight = BossFightWidgetSet(self.__controls, self.__render, Logger.instance(), self,
                                               self.__continue_explore, self.__popup_history.show,
                                               self.__save_data.check_unlocks, self._switch_to_menu,
                                               lambda: self.__map_manager.trigger_event(MapConfig.done_event_id()))
        self.__riddle = RiddleWidgetSet(self.__controls, self.__render, Logger.instance(), self,
                                        self.__continue_explore, self.__popup_history.show,
                                        self.__save_data.check_unlocks)
        self.__challenge = ChallengeWidgetSet(self.__controls, self.__render, Logger.instance(), self,
                                              self.__continue_explore, self.__popup_history.show,
                                              self.__save_data.check_unlocks)

        widget_sets: List[MyWidgetSet] = [self.__training, self.__navigation, self.__explore,
                                          self.__fight, self.__boss_fight, self.__riddle, self.__challenge,
                                          self.__workbench, self.__fusion_circuit]
        # INIT KEYS
        # add the general keys to everything except Transition, Menu and Pause
        for widget_set in widget_sets:
            # let widgets overwrite the pause key (e.g., disable pausing)
            widget_set.add_key_command(self.__controls.get_keys(Keys.Pause), self.__pause_game, add_to_widgets=True,
                                       overwrite_widgets=False)
            widget_set.add_key_command(self.__controls.get_keys(Keys.PopupReopen), self.__popup_history.show,
                                       add_to_widgets=True, overwrite=False)

        # debugging keys
        for widget_set in (widget_sets + [self.__transition, self.__menu, self.__pause]):
            for widget in widget_set.get_widget_list():
                widget.widget.add_key_command(self.__controls.get_keys(Keys.PrintScreen), self._print_screen)
                widget.widget.add_key_command(self.__controls.get_keys(Keys.Render), lambda: self.__render(None))

        # special cheat keys
        self.__pause.get_main_widget().add_key_command(self.__controls.get_keys(Keys.CheatInput),
                                                       CheatConfig.cheat_input)
        self.__pause.get_main_widget().add_key_command(self.__controls.get_keys(Keys.CheatList), CheatConfig.cheat_list)

        # INIT STATE MACHINE
        self.__cur_widget_set: MyWidgetSet = self.__transition  # avoid None value
        self.__state_machine = QrogueCUI._StateMachine(self)
        self.__state_machine.change_state(QrogueCUI._State.Menu, self.__rm.seed)

        # MISC
        if Config.debugging():
            self.set_on_draw_update_func(Config.inc_frame_count)

        Logger.print_to_console("Ready!")  # notify player that the game is fully loaded

    def _get_absolute_grid_dimensions(self):
        # wrapper because somehow the result is not updated if we pass grid.get_dimensions_absolute() as callback
        return self._grid.get_dimensions_absolute()

    def _refresh_height_width(self) -> None:
        try:
            super(QrogueCUI, self)._refresh_height_width()
            rows, cols = self._grid.get_dimensions_absolute()
            min_rows, min_cols = PyCuiConfig.get_min_dimensions()

            if rows < min_rows or cols < min_cols:
                raise PyCuiConfig.OutOfBoundsError()

        except PyCuiConfig.OutOfBoundsError:
            print("[Qrogue] ERROR!")
            rows, cols = self._grid.get_dimensions_absolute()
            print(f"Current dimensions are ({rows}, {cols}) but at least ({PyCuiConfig.get_min_dimensions()}) is "
                  f"needed. You can fix this by increasing the window size (e.g., maximize it) or adapting the font.")
            print("If you are running it in Windows Powershell you can also try to press ALT+ENTER to change to "
                  "fullscreen mode.")
            raise PyCuiConfig.OutOfBoundsError

    @property
    def is_simulating(self) -> bool:
        return self.__simulator is not None

    def set_refresh_timeout(self, timeout: int):
        """

        :param timeout: timeout in ms or -1 for no timeout
        :return: None
        """
        self._refresh_timeout = timeout
        if self._stdscr is not None:
            # handle the special case of stopping the draw-thread
            if self._refresh_timeout < 0:
                # we have to redraw manually since most likely the last update wasn't recognized by the draw-thread yet
                # calls are taken directly from super._draw() while ignoring input handling and logging
                self._stdscr.erase()
                # Draw status/title bar, and all widgets. Selected widget will be bolded.
                self._draw_status_bars(self._stdscr, self._height, self._width)
                self._draw_widgets()
                # draw the popup if required (although we most likely won't use popups during automated actions)
                if self._popup is not None:
                    self._popup._draw()
                self._stdscr.refresh()

            # since _draw is only called once, we have to set the timeout manually for the screen
            self._stdscr.timeout(self._refresh_timeout)

    def start(self, level_name: Optional[str] = None) -> NewSaveData:
        self.__ow_key_logger.reinit(self.__rm.seed, "meta", self.__save_data.to_keylog_string())

        if self.__save_data.is_fresh_save:
            def knowledge_question(index: int):
                if index == 0:
                    GameplayConfig.set_newbie_mode()
                else:
                    GameplayConfig.set_experienced_mode()

            ConfirmationPopup.ask("WELCOME TO QROGUE!", HelpText.Welcome.text, knowledge_question,
                                  [WordingConfig.QUANTUM_NEWBIE, WordingConfig.QUANTUM_EXPERIENCED])

        self.__render([self.__cur_widget_set])

        # We don't want to handle accidental input on startup of the game (e.g., during play-testing this once closed
        # the introduction popup before it was even visible to the player) so we set our init_complete-flag to True
        # after a short delay. Needs to be in an extra thread so _handle_key_presses() can try to handle the accidental
        # input. Otherwise, the input queue would not be cleared and the problem only delayed.
        def call_me():
            time.sleep(QrogueCUI.__INIT_DELAY)
            self.__init_complete = True

        Thread(target=call_me).start()

        if level_name is not None:
            self.__map_manager.load_map(level_name)

        super(QrogueCUI, self).start()
        return self.__save_data

    def stop(self) -> None:
        self.__ow_key_logger.flush_if_useful()
        super().stop()

    def _conditional_saving(self) -> Tuple[bool, CommonInfos]:
        if self.is_simulating:
            return False, CommonInfos.NoSavingDuringSimulation

        return self.__save_data.save()

    def _set_simulator(self, simulator: GameSimulator, auto_scroll_transitions: bool = False,
                       stop_when_finished: bool = False):
        self.__stored_save = self.__save_data
        self.__save_data = NewSaveData(simulator.save_state)
        self.__simulator = simulator
        self.__rm = RandomManager.create_new(simulator.seed)
        self.__ow_key_logger = DummyKeyLogger()
        self.__key_logger = DummyKeyLogger()
        self.__auto_scroll_simulation_transitions = auto_scroll_transitions
        self.__stop_with_simulation_end = stop_when_finished
        simulator.set_controls(self.__controls)

        if simulator.simulates_over_world:
            if self.__menu.seed != simulator.seed:
                self.__menu.set_data(simulator.seed)
                Config.check_reachability("QrogueCUI._set_simulator() with different seed", raise_exception=True)

        if simulator.version == Config.version():
            title, text = simulator.version_alright()
        else:
            title, text = simulator.version_warning()

        Popup.message(title, text, reopen=False, overwrite=True)

    def _end_simulation(self, can_stop: bool = True):
        """

        Args:
            can_stop: whether the game is allowed to automatically stop itself or not

        Returns:

        """
        assert self.__stored_save is not None
        self.__save_data = self.__stored_save
        self.__simulator = None
        self.set_refresh_timeout(-1)  # stop automation   # todo: check if this might interfere with transitions?
        if can_stop and self.__stop_with_simulation_end:
            self.stop()

    def _ready_for_input(self, key_pressed: int, gameplay: bool = True) -> bool:
        if self.__last_key != key_pressed:
            self.__last_key = key_pressed
            return True

        if gameplay:
            pause = GameplayConfig.get_option_value(Options.gameplay_key_pause, convert=True)
        else:
            pause = GameplayConfig.get_option_value(Options.simulation_key_pause, convert=True)
        now_time = time.time()
        if now_time - self.__last_input >= pause:
            self.__last_input = now_time
            return True
        return False

    def _handle_key_presses(self, key_pressed):
        # ignore all input before we completed initialization
        if not self.__init_complete: return

        if key_pressed == 0:  # skips the "empty" key press during initialization
            return
        if key_pressed == PyCuiConfig.KEY_CTRL_Q:
            super(QrogueCUI, self)._handle_key_presses(PyCuiConfig.KEY_ESCAPE)
        if self.__simulator is None:
            if self._ready_for_input(key_pressed, gameplay=True):
                if key_pressed == PyCuiConfig.KEY_ESCAPE:
                    pass  # ignore ESC because this makes you leave the CUI
                else:
                    if GameplayConfig.get_option_value(Options.log_keys) and not self.is_simulating:
                        self.__key_logger.log(self.__controls, key_pressed)
                        self.__ow_key_logger.log(self.__controls, key_pressed)
                    super(QrogueCUI, self)._handle_key_presses(key_pressed)
        elif key_pressed in self.__controls.get_keys(Keys.StopSimulator):
            Popup.message("Simulator", "stopped Simulator", reopen=False, overwrite=True)
            self._end_simulation(can_stop=False)  # since the player stopped the simulation
        else:
            if self._ready_for_input(key_pressed, gameplay=False):
                key = self.__simulator.next()
                if key is None:
                    Popup.message("Simulator", "finished", reopen=False, overwrite=True)
                    self._end_simulation()
                else:
                    super(QrogueCUI, self)._handle_key_presses(key)

    def _cycle_widgets(self, reverse: bool = False) -> None:
        pass  # this is neither needed nor allowed in Qrogue

    def _initialize_widget_renderer(self):
        """Function that creates the renderer object that will draw each widget
        """
        if self._renderer is None:
            self._renderer = MultiColorRenderer(self, self._stdscr, self._logger)
        super(QrogueCUI, self)._initialize_widget_renderer()

    def move_focus(self, widget: WidgetWrapper, auto_press_buttons: bool = True) -> None:
        if widget is None:
            Logger.instance().throw(Exception("Widget to focus is None!"))
        if isinstance(widget, PyCuiConfig.PyCuiWidget):
            super(QrogueCUI, self).move_focus(widget, auto_press_buttons)
        else:
            Logger.instance().throw(Exception(
                f"Non-PyCUI widget used in renderer! CurWidgetSet = {self.__cur_widget_set}"))

    def _print_screen(self) -> None:
        text = ""
        for my_widget in self.__cur_widget_set.get_widget_list():
            text += str(my_widget) + "\n"
            text += my_widget.widget.get_title()
            text += "\n"
        PathConfig.new_screen_print(text)

    def apply_widget_set(self, new_widget_set: MyWidgetSet) -> None:
        new_widget_set.reset()
        super().apply_widget_set(new_widget_set)
        self.__cur_widget_set = new_widget_set
        self.move_focus(self.__cur_widget_set.get_main_widget(), auto_press_buttons=False)
        self.__cur_widget_set.render()

    def show_message_popup(self, title: str, text: str, color: int = PyCuiColors.WHITE_ON_BLACK) -> None:
        # todo: seems like setting the focused widgets here is no longer needed? but further testing required
        self.__focused_widget = self.get_selected_widget()
        super(QrogueCUI, self).show_message_popup(title, text, color)

    def show_error_popup(self, title: str, text: str) -> None:
        # todo: seems like setting the focused widgets here is no longer needed? but further testing required
        self.__focused_widget = self.get_selected_widget()
        super(QrogueCUI, self).show_error_popup(title, text)

    def __show_message_popup(self, title: str, text: str, position: int, color: int,
                             dimensions: Optional[Tuple[int, int]] = None, reopen: Optional[bool] = None,
                             padding_x: Optional[int] = None, importance: Optional[PopupConfig.Importance] = None) \
            -> None:
        if reopen is None:
            reopen = False
        if importance is None:
            importance = PopupConfig.Importance.Undefined
        self.__focused_widget = self.get_selected_widget()
        self._popup = MultilinePopup(self, title, text, color, self._renderer, self._logger, self.__controls,
                                     pos=position, dimensions=dimensions,
                                     situational_callback=(self.__popup_history.back, self.__popup_history.forth),
                                     padding_x=padding_x)
        self.__popup_history.add(self._popup, is_permanent=reopen)

        # immediately close the popup if we want to ignore messages (they are still added to the history if important)
        if CheatConfig.ignore_dialogue(importance): self.close_popup()

    def __show_confirmation_popup(self, title: str, text: str, color: int, callback: Callable[[int], None],
                                  answers: Optional[List[str]]):
        self.__focused_widget = self.get_selected_widget()
        self._popup = MultilinePopup(self, title, text, color, self._renderer, self._logger, self.__controls, callback,
                                     answers)

    def __show_cheat_info_popup(self, title: str, text: str, position: int, color: int):
        self.__show_message_popup(title, text, position, color, importance=PopupConfig.Importance.Info)

    def __show_input_popup(self, title: str, color: int, callback: Callable[[str], None]) -> None:
        self.__focused_widget = self.get_selected_widget()
        self._popup = popups.TextBoxPopup(self, title, color, callback, self._renderer, False, self._logger)

    def close_popup(self) -> None:
        self.__popup_history.resolve()
        super(QrogueCUI, self).close_popup()
        self.move_focus(self.__focused_widget)
        Popup.on_close()

    def __general_continue(self):
        self.__state_machine.change_state(self.__state_machine.prev_state, None)

    def _switch_to_menu(self, data=None) -> None:
        if self.__key_logger and self.__key_logger.is_initialized:
            self.__key_logger.flush_if_useful()
            self.__key_logger.set_active(False)  # stop logging until we started a level again
        if data and data != self.__rm.seed:
            Config.check_reachability("switch_to_menu's seed setting")
            seed = data  # todo: I'm not sure if it is allowed to set a different seed than self.__rm.seed here
        else:
            seed = self.__rm.seed
        self.__menu.set_data(seed)
        self.apply_widget_set(self.__menu)

    def __show_level_select(self):
        self.__state_machine.change_state(QrogueCUI._State.LevelSelect, None)

    def _switch_to_level_select(self, _=None) -> None:
        self.__level_select.reinit()
        self.apply_widget_set(self.__level_select)
        if not self.__save_data.check_unlocks(Unlocks.VisitedLevelSelection):
            Popup.show_help(HelpText.LevelSelection)
            self.__save_data.unlock(Unlocks.VisitedLevelSelection)

    def __show_screen_check(self):
        self.__state_machine.change_state(QrogueCUI._State.ScreenCheck, None)

    def _switch_to_screen_check(self, data=None) -> None:
        self.apply_widget_set(self.__screen_check)

    def __start_playing(self):
        self.__map_manager.load_first_uncleared_map()

    def __start_expedition(self):
        Popup.error("Expeditions not yet unlocked!")

    def _switch_to_training(self, data=None):
        if data:
            robot, enemy = data
            self.__training.set_data(robot, enemy, False)
        self.apply_widget_set(self.__training)

    def __show_workbench(self):
        self.__state_machine.change_state(QrogueCUI._State.Workbench, None)

    def _switch_to_workbench(self, _=None):
        # no data parameter needed
        self.__workbench.reinit()
        self.apply_widget_set(self.__workbench)
        if not self.__save_data.check_unlocks(Unlocks.VisitedWorkbench):
            Popup.show_help(HelpText.Workbench)
            self.__save_data.unlock(Unlocks.VisitedWorkbench)

    def __show_fusion_circuit(self, gate_list: List[Instruction]):
        self.__state_machine.change_state(QrogueCUI._State.FusionCircuit, gate_list)

    def _switch_to_fusion_circuit(self, gate_list: List[Instruction]):
        self.__fusion_circuit.set_data(FusionBot(2, gate_list), FusionTarget(2), None)
        self.apply_widget_set(self.__fusion_circuit)

    def __start_level(self, level: Map) -> None:
        Logger.instance().info(f"Starting level {level.internal_name} with seed={level.seed}.", from_pycui=False)
        # reset in-level popup data
        self.__popup_history.reset()
        Popup.reset_queue()

        self.__map_manager.on_level_start()
        # store the level's seed and save state at the time of playing to the key logger
        self.__key_logger.reinit(level.seed, level.internal_name, self.__save_data.to_keylog_string())
        self.__ow_key_logger.level_start(level.internal_name)

        # reset the score at the start of each level
        level.robot.reset_score()   # needed because the player can restart levels to try again

        self.__pause.set_data(level.robot, level.name, None)
        self.__state_machine.change_state(QrogueCUI._State.Explore, level)

    def __game_over(self) -> None:
        def callback(confirmed: int):
            if confirmed == 0:
                self.__map_manager.reload()
            elif confirmed == 1:
                self.__state_machine.change_state(QrogueCUI._State.Menu, None)

        if self.__map_manager.is_in_level:
            text = "Do you want to retry the level?"
        elif self.__map_manager.is_in_expedition:
            text = "Do you want to retry the expedition?"
        else:
            text = "Do you want to retry?"
            Logger.instance().warn("__game_over() called without being in a level or expedition", from_pycui=False)
        ConfirmationPopup.ask(Config.system_name(), text, callback, ["Retry", "Back to Main Menu"])

    def __start_fight(self, robot: Robot, enemy: Enemy, direction: Direction) -> None:
        self.__state_machine.change_state(QrogueCUI._State.Fight, (robot, enemy))

    def __start_boss_fight(self, robot: Robot, boss: Boss, direction: Direction):
        self.__state_machine.change_state(QrogueCUI._State.BossFight, (robot, boss))

    def _switch_to_pause(self, data=None) -> None:
        self.apply_widget_set(self.__pause)

    def __pause_game(self) -> None:
        self.__state_machine.change_state(QrogueCUI._State.Pause, None)

    def _switch_to_explore(self, data: Optional[Union[Map, Tuple[Optional[Map], Optional[bool]]]]) -> None:
        if data is not None:
            if isinstance(data, Map):
                self.__explore.set_data(data)
            else:
                map_, undo_last_move = data
                if map_ is not None:
                    self.__explore.set_data(map_)
                if undo_last_move:
                    self.__explore.undo_last_move()
        self.apply_widget_set(self.__explore)
        self.__explore.try_to_start_map()

    def __continue_explore(self, undo_last_move: bool = False) -> None:
        self.__state_machine.change_state(QrogueCUI._State.Explore, (None, undo_last_move))

    def _switch_to_fight(self, data) -> None:
        if data is not None:
            robot = data[0]
            enemy = data[1]
            self.__fight.set_data(robot, enemy, self.__map_manager.show_individual_qubits)
        self.apply_widget_set(self.__fight)

    def _switch_to_boss_fight(self, data) -> None:
        if data is not None:
            robot = data[0]
            boss = data[1]
            self.__boss_fight.set_data(robot, boss, self.__map_manager.show_individual_qubits)
        self.apply_widget_set(self.__boss_fight)

    def __open_riddle(self, robot: Robot, riddle: Riddle):
        self.__state_machine.change_state(QrogueCUI._State.Riddle, (robot, riddle))

    def _switch_to_riddle(self, data) -> None:
        if data is not None:
            robot = data[0]
            riddle = data[1]
            self.__riddle.set_data(robot, riddle, self.__map_manager.show_individual_qubits)
        self.apply_widget_set(self.__riddle)

    def __open_challenge(self, robot: Robot, challenge: Challenge):
        self.__state_machine.change_state(QrogueCUI._State.Challenge, (robot, challenge))

    def _switch_to_challenge(self, data) -> None:
        if data is not None:
            robot = data[0]
            challenge = data[1]
            self.__challenge.set_data(robot, challenge, self.__map_manager.show_individual_qubits)
        self.apply_widget_set(self.__challenge)

    def _execute_transition(self, text_scrolls: List[TransitionWidgetSet.TextScroll], next_state: "QrogueCUI._State",
                            next_data: Any, additional_callback: Optional[Callable] = None):
        def callback():
            self.__state_machine.change_state(next_state, next_data)
            if additional_callback is not None:
                additional_callback()

        self.__state_machine.change_state(QrogueCUI._State.Transition, (text_scrolls, callback))

    def _switch_to_transition(self, data) -> None:
        if len(data) == 2:
            texts, continue_ = data
            self.__transition.set_data(texts, continue_)
        else:
            texts, continue_, auto_scroll = data
            self.__transition.set_data(texts, continue_, auto_scroll)
        self.apply_widget_set(self.__transition)

    def __render(self, renderables: Optional[List[Renderable]]):
        if renderables is None:
            renderables = self.__cur_widget_set.get_widget_list()
        for r in renderables:
            r.render()



import math
from abc import ABC, abstractmethod
from typing import List, Any, Callable, Tuple, Optional, Dict, Union, Iterator

from py_cui import ColorRule
from py_cui.widget_set import WidgetSet
from py_cui.widgets import BlockLabel

from qrogue.game.logic.actors import Robot
from qrogue.game.logic.base import StateVector, CircuitMatrix
from qrogue.game.logic.collectibles import Instruction
from qrogue.game.world.map import Map
from qrogue.game.world.navigation import Direction
from qrogue.graphics.popups import Popup
from qrogue.graphics.rendering import ColorRules
from qrogue.graphics.widget_base import WidgetWrapper
from qrogue.util import Controls, Keys, Logger, Options, Config, ColorConfig, HudConfig, GameplayConfig, \
    QuantumSimulationConfig, ErrorConfig
from qrogue.util.util_functions import align_string, int_to_fixed_len_str
from .renderable import Renderable


class MyBaseWidget(BlockLabel, WidgetWrapper):
    def __init__(self, wid, title, grid, row, column, row_span, column_span, padx, pady, center, logger):
        super().__init__(wid, title, grid, row, column, row_span, column_span, padx, pady, center, logger)

    def get_pos(self) -> Tuple[int, int]:
        return self._column, self._row

    def get_abs_pos(self) -> Tuple[int, int]:
        return self._start_x, self._start_y

    def get_size(self) -> Tuple[int, int]:
        return self._column_span, self._row_span

    def get_abs_size(self) -> Tuple[int, int]:
        return self._stop_x - self._start_x, self._stop_y - self._start_y

    def is_selected(self) -> bool:
        return super(MyBaseWidget, self).is_selected()

    def reposition(self, row: int = None, column: int = None, row_span: int = None, column_span: int = None):
        if row:
            self._row = row
        if column:
            self._column = column
        if row_span:
            self._row_span = row_span
        if column_span:
            self._column_span = column_span

    def set_title(self, title: str) -> None:
        # cannot easily use the original function to draw the border because the renderer draws it directly
        if self._draw_border:
            title = "-" * self._width + "\n" + title
        super(MyBaseWidget, self).set_title(title)

    def get_title(self) -> str:
        return super(MyBaseWidget, self).get_title()

    def add_text_color_rule(self, regex: str, color: int, rule_type: str, match_type: str = 'line',
                            region: List[int] = None, include_whitespace: bool = False, selected_color=None) \
            -> None:
        if region is None:
            region = [0, 1]
        super(MyBaseWidget, self).add_text_color_rule(regex, color, rule_type, match_type, region, include_whitespace,
                                                      selected_color)

    def reset_text_color_rules(self) -> None:
        self._text_color_rules.clear()

    def activate_individual_coloring(self):
        regex = ColorConfig.REGEX_TEXT_HIGHLIGHT
        self._text_color_rules.append(
            ColorRule(f"{regex}.*?{regex}", 0, 0, "contains", "regex", [0, 1], False, Logger.instance())
        )

    def add_key_command(self, keys: List[int], command: Callable[[], Any], overwrite: bool = True) -> Any:
        for key in keys:
            if overwrite or key not in self._key_commands:
                super(MyBaseWidget, self).add_key_command(key, command)


class MyMultiWidget(WidgetWrapper):
    @staticmethod
    def get_title_separator() -> str:
        return ">$%<"

    def __init__(self, widgets: List[WidgetWrapper]):
        assert len(widgets) > 0, "Emtpy MultiWidget created!"

        self.__widgets = widgets

        # the minimal (left-top most) position is where this widget starts (like casting a rectangle around all widgets)
        x, y = self.__widgets[0].get_pos()
        ax, ay = self.__widgets[0].get_abs_pos()
        for i in range(1, len(self.__widgets)):
            w = self.__widgets[i]
            # if logical position is smaller, than also absolute is
            wx, wy = w.get_pos()
            if wx < x:
                ax = w.get_abs_pos()[0]
                x = wx
            if wy < y:
                ay = w.get_abs_pos()[1]
                y = wy
        self.__pos = x, y
        self.__abs_pos = ax, ay

        widths: Dict[int, int] = {}  # find out width of the longest row
        heights: Dict[int, int] = {}  # and height of biggest column
        abs_widths: Dict[int, int] = {}
        abs_heights: Dict[int, int] = {}
        for w in self.__widgets:
            col, row = w.get_pos()
            width, height = w.get_size()
            if row not in widths or col + width > widths[row]:
                widths[row] = col + width
                abs_widths[row] = w.get_abs_pos()[0] + w.get_abs_size()[0]
            if col not in heights or row + height > heights[col]:
                heights[col] = row + height
                abs_heights[col] = w.get_abs_pos()[1] + w.get_abs_size()[1]
        self.__size = max(widths) - x, max(heights) - y
        self.__abs_size = max(abs_widths) - ax, max(abs_heights) - ay

    def get_pos(self) -> Tuple[int, int]:
        """
        Column of the left most widget, row of the top most widget. There will not necessarily be a real widget at
        exactly this position.
        :return: x, y / column, row
        """
        return self.__pos

    def get_abs_pos(self) -> Tuple[int, int]:
        return self.__abs_pos

    def get_size(self) -> Tuple[int, int]:
        """
        Width of the widest row and height of the highest column. You can imagine it like fitting the smallest possible
        rectangle around all its widgets.
        :return: width, height
        """
        return self.__size

    def get_abs_size(self) -> Tuple[int, int]:
        return self.__abs_size

    def is_selected(self) -> bool:
        """

        :return: True if one of its widgets is selected
        """
        for w in self.__widgets:
            if w.is_selected():
                return True
        return False

    def reposition(self, row: int = None, column: int = None, row_span: int = None, column_span: int = None):
        if column is None:
            col_diff = None
        else:
            col_diff = column - self.__pos[0]
        if row is None:
            row_diff = None
        else:
            row_diff = row - self.__pos[1]

        if column_span is None:
            width_diff = None
        else:
            width_diff = column_span - self.__size[0]
        if row_span is None:
            height_diff = None
        else:
            height_diff = row_span - self.__size[1]

        old_width, old_height = self.__size
        width_changes = 0
        height_changes = 0
        for i, w in enumerate(self.__widgets):
            w_col, w_row = w.get_pos()
            w_width, w_height = w.get_size()

            new_row = None
            if row_diff is not None:
                new_row = w_row + row_diff

            new_col = None
            if col_diff is not None:
                new_col = w_col + col_diff

            new_row_span = None
            if height_diff is not None:
                h_mul = height_diff * w_height / old_height  # try to keep the same widget_height / whole_height ratio
                height_change = round(w_height * h_mul)
                height_changes += height_change

                if i < len(self.__widgets) - 1 and height_changes < height_diff:
                    new_row_span = w_height + height_change
                else:
                    # give the widget the remaining space
                    new_row_span = height_diff - (height_changes - height_change)

            new_column_span = None
            if width_diff is not None:
                w_mul = width_diff * w_width / old_width  # try to keep the same widget_width / whole_width ratio
                width_change = round(w_width * w_mul)
                width_changes += width_change
                if width_changes < width_diff:
                    new_column_span = w_width + width_change

            # todo definitely needs detailed testing
            w.reposition(row=new_row, column=new_col, row_span=new_row_span, column_span=new_column_span)

        # update to the new values
        self.__pos = column, row
        self.__size = column_span, row_span

    def set_title(self, title: str) -> None:
        self.set_titles(title.split(MyMultiWidget.get_title_separator()))

    def set_titles(self, titles: List[str]):
        for i, t in enumerate(titles):
            if i < len(self.__widgets):
                self.__widgets[i].set_title(t)

    def get_title(self) -> str:
        titles = self.get_titles()
        return MyMultiWidget.get_title_separator().join(titles)

    def get_titles(self) -> List[str]:
        titles = []
        for w in self.__widgets:
            titles.append(w.get_title())
        return titles

    def add_text_color_rule(self, regex: str, color: int, rule_type: str, match_type: str = 'line',
                            region: List[int] = None, include_whitespace: bool = False, selected_color=None) -> None:
        """
        Applies the color rule to all of its widgets

        :param regex:
        :param color:
        :param rule_type:
        :param match_type:
        :param region:
        :param include_whitespace:
        :param selected_color:
        :return:
        """
        if region is None:
            region = [0, 1]

        for w in self.__widgets:
            w.add_text_color_rule(regex, color, rule_type, match_type, region, include_whitespace, selected_color)

    def reset_text_color_rules(self) -> None:
        for w in self.__widgets:
            w.reset_text_color_rules()

    def activate_individual_coloring(self):
        for w in self.__widgets:
            w.activate_individual_coloring()

    def add_key_command(self, keys: List[int], command: Callable[[], Any], overwrite: bool = True) -> Any:
        for w in self.__widgets:
            w.add_key_command(keys, command, overwrite)

    def toggle_border(self):
        for w in self.__widgets:
            w.toggle_border()


class Widget(Renderable, ABC):
    __MOVE_FOCUS: Optional[Callable[[WidgetWrapper, WidgetSet], None]] = None

    @staticmethod
    def set_move_focus_callback(move_focus: Callable[[WidgetWrapper, WidgetSet], None]):
        Widget.__MOVE_FOCUS = move_focus

    @staticmethod
    def move_focus(widget: "Widget", widget_set: WidgetSet):
        if Widget.__MOVE_FOCUS:
            Widget.__MOVE_FOCUS(widget.widget, widget_set)

    def __init__(self, widget: WidgetWrapper):
        self.__widget = widget

    @property
    def widget(self) -> WidgetWrapper:
        return self.__widget

    @abstractmethod
    def set_data(self, data) -> None:
        pass

    @abstractmethod
    def render(self) -> None:
        pass

    @abstractmethod
    def render_reset(self) -> None:
        pass


class SimpleWidget(Widget):
    def __init__(self, widget: WidgetWrapper, initial_text: str = ""):
        super().__init__(widget)
        self.__text = initial_text

    def set_data(self, data: str) -> None:
        self.__text = str(data)

    def render(self) -> None:
        self.widget.set_title(self.__text)

    def render_reset(self, reset_text: bool = True) -> None:
        if reset_text: self.__text = ""
        self.widget.set_title("")

    def __str__(self):
        return f"SimpleWidget(\"{self.__text}\")"


class HudWidget(Widget):
    __MAP_NAME = ""
    __SCORE_LENGTH = 5

    def __init__(self, widget: MyMultiWidget):
        super().__init__(widget)
        self.__robot = None
        self.__map_name = None
        self.__details = None
        self.__render_duration = None

    def set_data(self, data: Tuple[Robot, Optional[str], Optional[str]]) -> None:
        """
        :param data: Tuple of up to three data-items
                - data[0]: Robot
                - data[1]: name of the current map
                - data[2]: text for the situational HUD
        """
        self.__robot = data[0]
        if data[1] is not None:
            HudWidget.__MAP_NAME = data[1]
        if data[2] is not None:
            self.__details = data[2]
        else:
            self.__details = ""

        self.__map_name = HudWidget.__MAP_NAME

    def update_situational(self, data: str):
        self.__details = data

    def reset_data(self) -> None:
        self.__robot = None
        self.__map_name = None
        self.__details = None

    def update_render_duration(self, duration: float, force_render: bool = True):
        """
        Arguments:
            duration: time needed to render the screen in microseconds
            force_render: whether to immediately render or not
        """
        if Config.debugging():
            self.__render_duration = duration / 1000  # show it in milliseconds
            if force_render: self.render()

    def render(self) -> None:
        text = ""
        if HudConfig.ShowMapName and self.__map_name:
            text += f"{self.__map_name}\n"
        if self.__robot:
            if HudConfig.ShowScore:
                text += f"Score: {int_to_fixed_len_str(self.__robot.score, length=HudWidget.__SCORE_LENGTH)}  "
            if HudConfig.ShowEnergy:
                text += f"Energy: {self.__robot.cur_energy} / {self.__robot.max_energy}   \t"
            if HudConfig.ShowKeys:
                text += f"{self.__robot.key_count()} keys  \t"
        if HudConfig.ShowFPS and self.__render_duration:
            text += f"\t\t{self.__render_duration:.2f} ms"

        if Config.debugging():
            self.widget.set_title(f"{text}{MyMultiWidget.get_title_separator()}{self.__details}"
                                  f"{MyMultiWidget.get_title_separator()}Debug\n{Config.frame_count()}")
        else:
            self.widget.set_title(f"{text}{MyMultiWidget.get_title_separator()}{self.__details}")

    def render_reset(self) -> None:
        self.widget.set_title("")


class CircuitWidget(Widget):
    class _PlaceHolderData:
        def __init__(self, gate: Optional[Instruction], pos: int = -1, qubit: int = 0):
            self.gate = gate
            self.pos = pos
            self.qubit = qubit

        def resolve(self) -> Tuple[int, int]:
            return self.pos, self.qubit

        def can_change_position(self) -> bool:
            return self.gate is None or self.gate.no_qubits_specified

        def is_valid_qubit(self, qubit: int) -> bool:
            return self.gate is None or self.gate.can_use_qubit(qubit)

        def is_valid_pos(self, pos: int, robot: Robot) -> bool:
            # if gate is None we search for an occupied position (gate_used_at(pos) is not None)
            # if gate is not None we search for a free position (gate_used_at(pos) is None)
            # hence this xor condition
            return (self.gate is None) != (robot.gate_used_at(pos) is None)

        def place(self) -> bool:
            """

            :return: True if more qubits need to be placed, False otherwise
            """
            if self.gate is not None and self.gate.use_qubit(self.qubit):
                if self.qubit > 0:
                    self.qubit -= 1
                else:
                    self.qubit += 1
                return True
            return False

        def __iter__(self) -> Iterator[Union[Optional[Instruction], int, int]]:
            return iter((self.gate, self.pos, self.qubit))

    def __init__(self, widget: WidgetWrapper, controls: Controls):
        super().__init__(widget)
        self.__robot: Optional[Robot] = None
        self.__input: Optional[StateVector] = None
        self.__target: Optional[StateVector] = None
        self.__place_holder_data: Optional[CircuitWidget._PlaceHolderData] = None

        widget.add_key_command(controls.get_keys(Keys.SelectionUp), self.__move_up)
        widget.add_key_command(controls.get_keys(Keys.SelectionRight), self.__move_right)
        widget.add_key_command(controls.get_keys(Keys.SelectionDown), self.__move_down)
        widget.add_key_command(controls.get_keys(Keys.SelectionLeft), self.__move_left)

        widget.activate_individual_coloring()
        ColorRules.apply_circuit_rules(widget)

    def __change_position(self, right: bool) -> bool:
        def go_right(position: int) -> Optional[int]:
            if position + 1 >= self.__robot.circuit_space:
                return None
            return position + 1

        def go_left(position: int) -> Optional[int]:
            if position - 1 < 0:
                return None
            return position - 1

        if right:
            pos = go_right(self.__place_holder_data.pos)
        else:
            pos = go_left(self.__place_holder_data.pos)
        if pos is None:
            return False

        # only if we are currently removing a gate or if implicit removal is not allowed we have to check for for other
        # gates
        if self.__place_holder_data.gate is None or \
                not GameplayConfig.get_option_value(Options.allow_implicit_removal, convert=True):
            # go on until we find a valid position for the gate
            while not self.__place_holder_data.is_valid_pos(pos, self.__robot):
                if right:
                    pos = go_right(pos)
                else:
                    pos = go_left(pos)
                if pos is None:
                    return False

        self.__place_holder_data.pos = pos
        self.render()
        return True

    def __move_up(self):
        if self.__place_holder_data:
            qubit = self.__place_holder_data.qubit + 1
            while qubit < self.__robot.num_of_qubits:
                if self.__place_holder_data.is_valid_qubit(qubit):
                    self.__place_holder_data.qubit = qubit
                    self.render()
                    return
                qubit += 1

    def __move_right(self):
        if self.__place_holder_data:
            if self.__place_holder_data.can_change_position():
                self.__change_position(True)

    def __move_down(self):
        if self.__place_holder_data:
            qubit = self.__place_holder_data.qubit - 1
            while qubit >= 0:
                if self.__place_holder_data.is_valid_qubit(qubit):
                    self.__place_holder_data.qubit = qubit
                    self.render()
                    return
                qubit -= 1

    def __move_left(self):
        if self.__place_holder_data:
            if self.__place_holder_data.can_change_position():
                self.__change_position(False)

    def abort_placement(self):
        if self.__place_holder_data is not None:
            if self.__place_holder_data.gate is not None:
                self.__place_holder_data.gate.reset()
            self.__place_holder_data = None
            self.render()
            # todo

    def start_gate_placement(self, gate: Optional[Instruction], pos: int = -1, qubit: int = 0):
        self.__place_holder_data = self._PlaceHolderData(gate, pos, qubit)
        if pos < 0 or self.__robot.circuit_space <= pos:
            # if we're currently not removing a gate and implicit removal is allowed we can definitely start at any
            # position
            if self.__place_holder_data.gate is not None and \
                    GameplayConfig.get_option_value(Options.allow_implicit_removal, convert=True):
                self.__place_holder_data.pos = 0
            else:
                for i in range(self.__robot.circuit_space):
                    if self.__place_holder_data.is_valid_pos(i, self.__robot):
                        self.__place_holder_data.pos = i
                        break

    def place_gate(self) -> Tuple[bool, Optional[Instruction]]:
        """

        :return: True if gate is fully placed, False otherwise (e.g. more qubits need to be placed)
        """
        if self.__place_holder_data:
            if self.__place_holder_data.gate is None:
                # remove the instruction
                gate = self.__robot.gate_used_at(self.__place_holder_data.pos)
                if gate is not None:
                    self.__robot.remove_instruction(gate)
                    self.__place_holder_data = None
                    self.render()
                    return True, None
            else:
                if self.__place_holder_data.gate.num_of_qubits > self.__robot.num_of_qubits:
                    Popup.error(ErrorConfig.qubit_overflow(self.__place_holder_data.gate.num_of_qubits,
                                                           self.__robot.num_of_qubits))
                    # return value doesn't really matter here, but we chose "True" to not continue placement
                    return True, None

                if self.__place_holder_data.place():
                    self.render()
                    return False, self.__place_holder_data.gate
                if self.__robot.use_instruction(self.__place_holder_data.gate, self.__place_holder_data.pos):
                    gate = self.__place_holder_data.gate
                    self.__place_holder_data = None
                    return True, gate
                Logger.instance().error("Place_Gate() did not work correctly", show=False, from_pycui=False)
        return False, None

    def set_data(self, data: Tuple[Robot, Optional[Tuple[StateVector, StateVector]]]) -> None:
        self.__robot, vectors = data
        if vectors is None:
            self.__input, self.__target = None, None
        else:
            self.__input, self.__target = vectors

    def render(self) -> None:
        if self.__robot is not None:
            gates: Dict[int, Instruction] = {}
            for gate in self.__robot.instructions:
                if gate.position is not None:
                    gates[gate.position] = gate
            circ_str = Instruction.circuit_to_string(self.__robot.num_of_qubits, self.__robot.circuit_space, gates,
                                                     self.__place_holder_data,
                                                     (self.__input, self.__robot.state_vector, self.__target))
            self.widget.set_title(circ_str)

    def render_reset(self) -> None:
        self.widget.set_title("")


class MapWidget(Widget):
    def __init__(self, widget: WidgetWrapper):
        super().__init__(widget)
        self.__map_started = False
        self.__map: Optional[Map] = None
        self.__backup = None

    def set_data(self, map_: Map) -> None:
        self.__map = map_
        self.__map_started = False

    def try_to_start_map(self):
        if not self.__map_started:
            if self.__map is None:
                Logger.instance().throw(RuntimeError("self.__map is None! Most likely this means that the map could "
                                                     "not be loaded."))
            self.__map.start()
            self.__map_started = True

    def render(self) -> None:
        if self.__map is not None:
            rows = self.__map.row_strings()
            # add robot
            x = self.__map.robot_pos.x
            y = self.__map.robot_pos.y
            rows[y] = rows[y][0:x] + self.__map.robot_img + rows[y][x + 1:]

            self.widget.set_title("\n".join(rows))

    def render_reset(self) -> None:
        self.__backup = self.widget.get_title().title()
        self.widget.set_title("")

    def move(self, direction: Direction) -> bool:
        return self.__map.move(direction)

    def undo_last_move(self) -> bool:
        if self.__map is not None:
            return self.__map.undo_last_move()
        return False


class StateVectorWidget(Widget):
    def __init__(self, widget: WidgetWrapper, headline: str):
        super().__init__(widget)
        self.__headline = headline
        self._stv_str_rep = None
        ColorRules.apply_heading_rules(widget)
        ColorRules.apply_qubit_config_rules(widget)

    @property
    def _headline(self) -> str:
        return f"~{self.__headline}~\n\n"

    @property
    def can_display_all_content(self) -> bool:
        width, height = self.widget.get_abs_size()
        content = self._stv_str_rep
        if len(content) > 0:
            lines = content.splitlines(keepends=False)
            max_line_len = max([len(line) for line in lines])
            return max_line_len <= width and len(lines) <= height
        else:
            return True

    def set_data(self, state_vector: StateVector) -> None:
        self._stv_str_rep = self._headline + state_vector.to_string()

    def render(self) -> None:
        if self._stv_str_rep:
            self.widget.set_title(self._stv_str_rep)

    def render_reset(self) -> None:
        self.widget.set_title("")


class InputStateVectorWidget(StateVectorWidget):
    def __init__(self, widget: WidgetWrapper, headline: str):
        super().__init__(widget, headline)

    def set_data(self, state_vector: StateVector) -> None:
        self._stv_str_rep = self._headline + state_vector.to_string()


class OutputStateVectorWidget(StateVectorWidget):
    def __init__(self, widget: WidgetWrapper, headline: str):
        super().__init__(widget, headline)
        widget.activate_individual_coloring()

    def set_data(self, state_vectors: Tuple[StateVector, StateVector], target_reached: bool = False) -> None:
        """

        Args:
            state_vectors: Tuple of the output statevector to display and its diff to a target statevector to color the
                            output correspondingly
            target_reached: _no longer in use_

        Returns: None

        """
        output_stv, diff_stv = state_vectors

        def wrap(skip_ket: bool):
            return [
                output_stv.wrap_in_qubit_conf(
                    i, coloring=True,
                    # check if diff is small enough
                    correct_amplitude=abs(diff_stv.at(i)) <= QuantumSimulationConfig.TOLERANCE,
                    skip_ket=skip_ket)
                for i in range(output_stv.size)  # do it for every qubit combination
            ]

        lines = wrap(skip_ket=False)
        # check if the content fits its widget
        max_line_len = max([len(line) for line in lines])
        width, _ = self.widget.get_abs_size()
        if max_line_len > width: lines = wrap(skip_ket=True)  # shrink content by removing ket

        self._stv_str_rep = self._headline + "\n".join(lines)


class TargetStateVectorWidget(StateVectorWidget):
    def __init__(self, widget: WidgetWrapper, headline: str):
        super().__init__(widget, headline)

    def set_data(self, state_vector: StateVector) -> None:
        def wrap(skip_ket: bool):  # wrap values according to TargetStv specifics (show percentages)
            return [
                state_vector.wrap_in_qubit_conf(
                    i, coloring=False, show_percentage=True,
                    skip_ket=skip_ket)
                for i in range(state_vector.size)  # do it for every qubit combination
            ]

        lines = wrap(skip_ket=False)
        # check if the content fits its widget
        max_line_len = max([len(line) for line in lines])
        width, _ = self.widget.get_abs_size()

        # "-2" is magic number found by trial and error that gives feasible results (perfect visual results are hard
        # since it also depends on font and other spacings); possible explanation: coloring of ket needs 2 chars on the
        # right end, meaning if they would be the only cut-off PyCUI could be smart enough to still color it
        if max_line_len > width - 2:
            lines = wrap(skip_ket=True)  # shrink content by removing ket
            # add whitespace so headline is not in the center but more above the amplitudes for better visuals
            split_index = self._headline.index("\n")
            headline = self._headline[:split_index] + " " * QuantumSimulationConfig.MAX_PERCENTAGE_SPACE \
                       + self._headline[split_index:]
            self._stv_str_rep = headline + "\n".join(lines)
        else:
            self._stv_str_rep = self._headline + "\n".join(lines)


class CircuitMatrixWidget(Widget):
    def __init__(self, widget: WidgetWrapper):
        super().__init__(widget)
        self.__matrix_str_rep = None
        ColorRules.apply_heading_rules(widget)
        ColorRules.apply_qubit_config_rules(widget)

    @property
    def can_display_all_content(self) -> bool:
        width, height = self.widget.get_abs_size()
        content = self.__matrix_str_rep
        if len(content) > 0:
            lines = content.splitlines(keepends=False)
            max_line_len = max([len(line) for line in lines])
            return max_line_len <= width and len(lines) <= height
        return True

    def set_data(self, matrix: CircuitMatrix) -> None:
        self.__matrix_str_rep = f"~Circuit Matrix~\n"
        if matrix.num_of_qubits > 3:
            self.__matrix_str_rep += "\n" * int(0.5 * matrix.size - 1)
            self.__matrix_str_rep += "Matrix is too big to be displayed!\n"
            # self.__matrix_str_rep += "But you can have a look at it by opening:\n"
            # self.__matrix_str_rep += " \"TODO\""        # todo create html file of current matrix?
        else:
            self.__matrix_str_rep += matrix.to_string()

    def render(self) -> None:
        if self.__matrix_str_rep is not None:
            self.widget.set_title(self.__matrix_str_rep)

    def render_reset(self) -> None:
        self.widget.set_title("")


class QubitInfoWidget(Widget):
    def __init__(self, widget: WidgetWrapper, left_aligned: bool = True):
        super(QubitInfoWidget, self).__init__(widget)
        self.__left_aligned = left_aligned
        self.__text = ""
        widget.add_text_color_rule("~.*~", ColorConfig.QUBIT_INFO_COLOR, 'contains', match_type='regex')

    def set_data(self, num_of_qubits: int) -> None:
        head = ""
        body = ""
        box_left = "|" + " " * 2
        box_right = " " * 2 + "|"
        if self.__left_aligned:
            head_range = range(num_of_qubits - 1, -1, -1)
        else:
            head_range = range(num_of_qubits)

        for i in head_range:
            head += f" q{i} "
        # skip 1 character left because of the additional ~ and 2 right because of the qubit index and ~
        head = box_left[:-1] + "~" + head[1:-1] + "~" + box_right[2:]

        for i in range(2 ** num_of_qubits):
            bin_num = bin(i)[2:]  # get rid of the '0b' at the beginning of the binary representation
            # add 0s to the beginning (left) by justifying the text to the right
            bin_num = bin_num.rjust(num_of_qubits, '0')
            row = "   ".join(bin_num)  # separate the digits in the string with spaces
            if not self.__left_aligned:
                row = row[::-1]  # [::-1] reverses the list so q0 is on the left
            body += box_left + row + box_right
            body += "\n"

        self.__text = head + "\n" + body
        self.widget.set_title(self.__text)

    def render(self) -> None:
        # self.widget.set_title(self.__text)
        pass

    def render_reset(self) -> None:
        self.widget.set_title("")


class SelectionWidget(Widget):
    __SELECTION_MARKER = "-> "
    __SEPARATOR = " " * len(__SELECTION_MARKER)

    @staticmethod
    def cancel_obj() -> Any:
        """
        A selection object that can be used globally to indicate a "Cancel" choice
        """
        return "-cancel-"

    @staticmethod
    def wrap_in_hotkey_str(options: List[str]) -> List[str]:
        if len(options) <= 1:
            return options  # no explicit hotkeys if there are not multiple options
        wrapped_options = []
        for i, option in enumerate(options):
            wrapped_options.append(SelectionWidget._wrap_in_hotkey_str(option, i))
        return wrapped_options

    @staticmethod
    def _wrap_in_hotkey_str(text: str, index: int) -> str:
        return f"[{index}] {text}"

    @staticmethod
    def _is_wrapped_in_hotkey_str(text: str, index: int) -> bool:
        return text.startswith(f"[{index}] ")

    def __init__(self, widget: WidgetWrapper, controls: Controls, columns: int = 1, is_second: bool = False,
                 stay_selected: bool = False, add_cancel_key: bool = True,
                 on_key_press: Optional[Callable[[Keys], None]] = None):
        """
        :param widget: the underlying widget
        :param controls: used Controls for key mapping
        :param columns: in how many columns the choices should be represented
        :param is_second: whether this SelectionWidget has a first "main" SelectionWidget that controls its content
        :param stay_selected: whether the selected choice should stay when focusing another widget or not
        :param add_cancel_key: whether a cancel functionality should be added (only works if there is a cancel_obj!)
        :param on_key_press: a method that is called after (i.e., the SelectionWidgets internal state already changed)
            a selection-changing key or action is pressed
        """
        super(SelectionWidget, self).__init__(widget)
        self.__columns = columns
        self.__is_second = is_second
        self.__stay_selected = stay_selected

        def okp(key: Keys):
            if on_key_press is not None: on_key_press(key)

        self.__on_key_press = okp
        self.__index = 0
        self.__choices: List[str] = []
        self.__choice_objects: List = []
        self.__callbacks: List[Union[Callable[[], Optional[bool]]], Callable[[int], Optional[bool]]] = []
        self.widget.add_text_color_rule(f"->", ColorConfig.SELECTION_COLOR, 'contains', match_type='regex')

        # init keys
        self.widget.add_key_command(controls.get_keys(Keys.SelectionUp), self._up)
        self.widget.add_key_command(controls.get_keys(Keys.SelectionRight), self._right)
        self.widget.add_key_command(controls.get_keys(Keys.SelectionDown), self._down)
        self.widget.add_key_command(controls.get_keys(Keys.SelectionLeft), self._left)

        if add_cancel_key:
            self.widget.add_key_command(controls.get_keys(Keys.Cancel), self.__jump_to_cancel, overwrite=False)

        # sadly cannot use a loop here because of how lambda expressions work the index would be the same for all calls
        # instead we use a list of indices to still be flexible without changing much code
        indices = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        hotkeys = [
            controls.get_keys(Keys.HotKey0),
            controls.get_keys(Keys.HotKey1),
            controls.get_keys(Keys.HotKey2),
            controls.get_keys(Keys.HotKey3),
            controls.get_keys(Keys.HotKey4),
            controls.get_keys(Keys.HotKey5),
            controls.get_keys(Keys.HotKey6),
            controls.get_keys(Keys.HotKey7),
            controls.get_keys(Keys.HotKey8),
            controls.get_keys(Keys.HotKey9),
        ]
        self.widget.add_key_command(hotkeys[indices[0]], lambda: self.__jump_to_index(indices[0]))
        self.widget.add_key_command(hotkeys[indices[1]], lambda: self.__jump_to_index(indices[1]))
        self.widget.add_key_command(hotkeys[indices[2]], lambda: self.__jump_to_index(indices[2]))
        self.widget.add_key_command(hotkeys[indices[3]], lambda: self.__jump_to_index(indices[3]))
        self.widget.add_key_command(hotkeys[indices[4]], lambda: self.__jump_to_index(indices[4]))
        self.widget.add_key_command(hotkeys[indices[5]], lambda: self.__jump_to_index(indices[5]))
        self.widget.add_key_command(hotkeys[indices[6]], lambda: self.__jump_to_index(indices[6]))
        self.widget.add_key_command(hotkeys[indices[7]], lambda: self.__jump_to_index(indices[7]))
        self.widget.add_key_command(hotkeys[indices[8]], lambda: self.__jump_to_index(indices[8]))
        self.widget.add_key_command(hotkeys[indices[9]], lambda: self.__jump_to_index(indices[9]))

    @property
    def columns(self) -> int:
        return self.__columns

    @property
    def rows(self) -> int:
        return math.ceil(self.num_of_choices / self.columns)

    @property
    def _width_of_last_row(self) -> int:
        return self.num_of_choices - self._index_of_row_start(self.num_of_choices - 1)  # - start of last row

    @property
    def num_of_choices(self) -> int:
        return len(self.__choices)

    @property
    def index(self) -> int:
        return self.__index

    @property
    def selected_object(self) -> Any:
        if self.__choice_objects is not None and self.index < len(self.__choice_objects):
            return self.__choice_objects[self.index]
        return None

    def _index_of_row_start(self, index: int):
        start_of_last_row = self.columns * (self.rows - 1)
        if index >= start_of_last_row:
            return start_of_last_row
        return index - (index % self.columns)

    def update_text(self, text: str, index: int):
        if 0 <= index < len(self.__choices):
            if SelectionWidget._is_wrapped_in_hotkey_str(self.__choices[index], index):
                self.__choices[index] = SelectionWidget._wrap_in_hotkey_str(text, index)
            else:
                self.__choices[index] = text

    def set_data(self, data: Union[
        Tuple[List[str], List[Callable[[], Optional[bool]]]],
        Tuple[List[str], List[Callable[[int], Optional[bool]]]],
        Tuple[Tuple[List[str], List[Any]], List[Callable[[], Optional[bool]]]],
        Tuple[Tuple[List[str], List[Any]], Callable[[int], Optional[bool]]],
        List[Tuple[str, Callable[[], Optional[bool]]]],
        List[Tuple[Tuple[str, Any], Callable[[], Optional[bool]]]],
    ]) -> None:
        """
        Arguments:
            data: selection text (optionally also selection objects) and corresponding action callback either as

                1) Tuple: List of str (=texts), List of Callable() (=actions),
                    E.g.: ([text1, text2], [action1(), action2()])

                2) Tuple: List of str (=texts), List of Callable(index) (=action),
                    E.g.: ([text1, text2], [action(i)])

                3) Tuple: Tuple of (List of str (=texts), List of Any (=objects)), List of Callable() (=actions),
                    E.g.: (([text1, text2], [obj1, obj2]), [action1, action2])

                4) Tuple: Tuple of (List of str (=texts), List of Any (=objects)), Callable(index) (=action),
                    E.g.: (([text1, text2], [obj1, obj2]), action(i))

                5) List: Tuple of (str (=text), Callable() (=action)),
                    E.g.: [(text1, action1), (text2, action2)]

                6) List: Tuple of (Tuple of (str (=text), Any (=object)), Callable() (=action)),
                    E.g.: [((text1, obj1), action1), ((text2, obj2), action2)]

        """
        assert len(data) >= 1, "set_data() called with empty data!"

        self.render_reset()

        if isinstance(data, list):
            self.__choices = []
            self.__callbacks = []
            for elem in data:
                if isinstance(elem[0], tuple):
                    text, obj = elem[0]
                    self.__choices.append(text)
                    self.__choice_objects.append(obj)
                else:
                    self.__choices.append(elem[0])
                self.__callbacks.append(elem[1])
        else:
            self.__choices, callbacks = data
            if isinstance(callbacks, list):
                self.__callbacks = callbacks
            else:
                self.__callbacks = [callbacks]

        if isinstance(self.__choices, tuple):
            self.__choices, self.__choice_objects = self.__choices

        if self.__columns > 1:
            choice_length = max([len(choice) for choice in self.__choices])
            for i in range(len(self.__choices)):
                self.__choices[i] = self.__choices[i].ljust(choice_length)

        for i, val in enumerate(self.__choices):
            if val is None:
                self.__choices[i] = "-Error: no text specified-"

    def render(self) -> None:
        rows = [""] * self.rows
        cur_row = 0
        for i in range(self.num_of_choices):
            if i == self.__index and (self.widget.is_selected() or self.__stay_selected):
                rows[cur_row] += SelectionWidget.__SELECTION_MARKER
            else:
                rows[cur_row] += SelectionWidget.__SEPARATOR
            rows[cur_row] += self.__choices[i]
            rows[cur_row] += " "
            if i % self.__columns == self.__columns - 1:
                cur_row += 1
            else:
                rows[cur_row] += SelectionWidget.__SEPARATOR

        if len(rows) > 0:  # simple validity check since some selections are dynamically created during runtime
            max_row_len = max([len(row) for row in rows])
            aligned_rows = [align_string(row, max_row_len) for row in rows]
            self.widget.set_title("\n".join(aligned_rows))

    def render_reset(self) -> None:
        self.widget.set_title("")
        self.__index = 0
        # This content reset is needed because changes to other widgets can cause this to render, making the reset
        # pointless if content is still set. Should this lead to any problems somewhere, a new flag-parameter can be
        # added.
        if self.__is_second:
            self.__choices = []
            self.__choice_objects = []
            self.__callbacks = []

    def validate_index(self) -> bool:
        if self.__index < 0:
            self.__index = 0
            return False
        if len(self.__choices) <= self.__index:
            self.__index = len(self.__choices) - 1
            return False
        return True

    def __single_next(self) -> None:
        self.__index += 1
        if self.__index >= self.num_of_choices:
            self.__index = 0

    def __single_prev(self) -> None:
        self.__index -= 1
        if self.__index < 0:
            self.__index = self.num_of_choices - 1

    def _up(self) -> None:
        if self.num_of_choices <= 1:
            return
        if self.num_of_choices <= self.__columns or self.__columns == 1:
            self.__single_prev()
        else:
            # special case for first line
            if self.__index < self.__columns:
                # provides either the very last element or (start of last row + column (== index if in first row))
                self.__index = min(self._index_of_row_start(self.num_of_choices - 1) + self.__index,
                                   self.num_of_choices - 1)
            else:
                self.__index -= self.__columns
        # only call if the key press changes something (e.g. more than 1 choice)
        self.__on_key_press(Keys.SelectionUp)
        self.render()

    def _right(self) -> None:
        if self.num_of_choices <= 1:
            return
        if self.__columns == 1 or self.num_of_choices <= self.__columns:
            self.__single_next()
        else:
            # special case if we are currently at the end of the last row
            if self.__index == self.num_of_choices - 1:
                self.__index = self._index_of_row_start(self.__index)
            # another special case if we are at the end of any other row
            elif self.__index % self.__columns == self.__columns - 1:
                self.__index -= (self.__columns - 1)
            else:
                self.__index += 1
        # only call if the key press changes something (e.g. more than 1 choice)
        self.__on_key_press(Keys.SelectionRight)
        self.render()

    def _down(self) -> None:
        if self.num_of_choices <= 1:
            return
        if self.num_of_choices <= self.__columns or self.__columns == 1:
            self.__single_next()
        else:
            # special case if we are currently in the last line
            if self.__index >= self._index_of_row_start(self.num_of_choices - 1):
                self.__index = self.__index % self.__columns
            else:
                self.__index = min(self.__index + self.__columns, self.num_of_choices - 1)
        # only call if the key press changes something (e.g. more than 1 choice)
        self.__on_key_press(Keys.SelectionDown)
        self.render()

    def _left(self) -> None:
        if self.num_of_choices <= 1:
            return
        if self.__columns == 1 or self.num_of_choices <= self.__columns:
            self.__single_prev()
        else:
            # special case if we are currently at the start of the last row
            if self.__index == self._index_of_row_start(self.num_of_choices - 1):
                self.__index = self.num_of_choices - 1
            # another special case if we are at the start of any other row
            elif self.__index % self.columns == 0:
                self.__index += self.__columns - 1
            else:
                self.__index -= 1
        # only call if the key press changes something (e.g. more than 1 choice)
        self.__on_key_press(Keys.SelectionLeft)
        self.render()

    def __jump_to_cancel(self):
        # try to find cancel_obj to jump to its index
        if len(self.__choice_objects) <= 0: return  # there is no cancel_obj

        # go from back to front because usually cancel is one of the last choices
        for i in range(self.num_of_choices, -1, -1):
            if len(self.__choice_objects) <= i: continue    # still out of bounds
            if self.__choice_objects[i] is SelectionWidget.cancel_obj():
                # we found cancel_obj
                self.__jump_to_index(i)

    def __jump_to_index(self, index: int):
        if index < 0:
            self.__index = 0
        elif self.num_of_choices <= index:
            self.__index = self.num_of_choices - 1
        else:
            self.__index = index
        self.__on_key_press(Keys.hotkey(index))
        self.render()

    def use(self) -> bool:
        """
        :return: True if the focus should move, False if the focus should stay in this SelectionWidget
        """
        # if only one callback is given, it needs the index as parameter
        if len(self.__callbacks) == 1 and self.num_of_choices > 1:
            ret = self.__callbacks[0](self.__index)
        else:
            if self.__index >= len(self.__callbacks):
                Logger.instance().throw(IndexError(f"Invalid index = {self.__index} for {self.__callbacks}. "
                                                   f"Text of choices: {self.__choices}"))
            ret = self.__callbacks[self.__index]()
        self.__on_key_press(Keys.Action)
        # return True for None to also move focus if nothing is returned
        return True if ret is None else ret


class HistoricWrapperWidget:
    def __init__(self, widgets: List[Widget], render_widgets: bool, save_initial_state: bool = True):
        self.__widgets = widgets
        self.__history: List[Tuple[str]] = []
        self.__index = -1

        if render_widgets:
            for widget in widgets:
                widget.render()  # update the visuals before potentially saving them
        if save_initial_state:
            self.save_state(rerender=True, force=True)

    @property
    def index(self) -> int:
        return self.__index

    @property
    def is_in_past(self) -> bool:
        return self.__index < len(self.__history) - 1

    @property
    def _cur_data(self) -> Optional[Tuple[str]]:
        if 0 <= self.__index < len(self.__history):
            return self.__history[self.__index]
        return None

    def save_state(self, rerender: bool = False, force: bool = False):
        if rerender:
            for widget in self.__widgets:
                widget.render()
        data = tuple([widget.widget.get_title() for widget in self.__widgets])

        # if force is False then data is not saved if it is equal to the latest data
        if force or len(self.__history) <= 0 or data != self.__history[-1]:
            self.__history.append(data)
            self.__index = len(self.__history) - 1  # index points to the last element

    def _back(self, render: bool):
        self.__index -= 1
        self.__index = max(self.__index, 0)
        if render: self.render()

    def _forth(self, render: bool):
        self.__index += 1
        self.__index = min(self.__index, len(self.__history) - 1)
        if render: self.render()

    def travel(self, forth: bool, render: bool = True):
        if forth:
            self._forth(render)
        else:
            self._back(render)

    def jump_to_present(self, render: bool = True):
        if self.is_in_past:
            self.__index = len(self.__history) - 1
            if render: self.render()

    def clear_history(self) -> None:
        self.__history.clear()
        self.__index = 0

    def render(self) -> None:
        if self._cur_data is not None:
            assert len(self._cur_data) == len(self.__widgets), "Length of data doesn't match widgets': " \
                                                               f"{len(self._cur_data)} != {len(self.__widgets)}"
            for i, widget in enumerate(self.__widgets):
                widget.widget.set_title(self._cur_data[i])
                # widget.render()



from typing import Tuple, Optional

from qrogue.game.logic.base import StateVector, CircuitMatrix
from qrogue.game.logic.collectibles import Collectible
from .target import Target


class FusionTarget(Target):
    def __init__(self, num_of_qubits: int):
        super().__init__(id_=0, target=StateVector.create_zero_state_vector(num_of_qubits), reward=None)

    def is_reached(self, state_vector: StateVector, circ_matrix: CircuitMatrix) -> Tuple[bool, Optional[Collectible]]:
        return False, None  # since we just mock a target, we can claim to never reach it



# Generated from D:/Workspaces/pycharm-workspace/Qrogue/qrogue/game/world/dungeon_generator/QrogueDungeon.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .QrogueDungeonParser import QrogueDungeonParser
else:
    from QrogueDungeonParser import QrogueDungeonParser

# This class defines a complete listener for a parse tree produced by QrogueDungeonParser.
class QrogueDungeonListener(ParseTreeListener):

    # Enter a parse tree produced by QrogueDungeonParser#start.
    def enterStart(self, ctx:QrogueDungeonParser.StartContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#start.
    def exitStart(self, ctx:QrogueDungeonParser.StartContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#meta.
    def enterMeta(self, ctx:QrogueDungeonParser.MetaContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#meta.
    def exitMeta(self, ctx:QrogueDungeonParser.MetaContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#robot.
    def enterRobot(self, ctx:QrogueDungeonParser.RobotContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#robot.
    def exitRobot(self, ctx:QrogueDungeonParser.RobotContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#room_content.
    def enterRoom_content(self, ctx:QrogueDungeonParser.Room_contentContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#room_content.
    def exitRoom_content(self, ctx:QrogueDungeonParser.Room_contentContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#r_row.
    def enterR_row(self, ctx:QrogueDungeonParser.R_rowContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#r_row.
    def exitR_row(self, ctx:QrogueDungeonParser.R_rowContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#tile.
    def enterTile(self, ctx:QrogueDungeonParser.TileContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#tile.
    def exitTile(self, ctx:QrogueDungeonParser.TileContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#tile_descriptor.
    def enterTile_descriptor(self, ctx:QrogueDungeonParser.Tile_descriptorContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#tile_descriptor.
    def exitTile_descriptor(self, ctx:QrogueDungeonParser.Tile_descriptorContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#t_descriptor.
    def enterT_descriptor(self, ctx:QrogueDungeonParser.T_descriptorContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#t_descriptor.
    def exitT_descriptor(self, ctx:QrogueDungeonParser.T_descriptorContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#trigger_descriptor.
    def enterTrigger_descriptor(self, ctx:QrogueDungeonParser.Trigger_descriptorContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#trigger_descriptor.
    def exitTrigger_descriptor(self, ctx:QrogueDungeonParser.Trigger_descriptorContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#teleport_descriptor.
    def enterTeleport_descriptor(self, ctx:QrogueDungeonParser.Teleport_descriptorContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#teleport_descriptor.
    def exitTeleport_descriptor(self, ctx:QrogueDungeonParser.Teleport_descriptorContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#message_descriptor.
    def enterMessage_descriptor(self, ctx:QrogueDungeonParser.Message_descriptorContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#message_descriptor.
    def exitMessage_descriptor(self, ctx:QrogueDungeonParser.Message_descriptorContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#collectible_descriptor.
    def enterCollectible_descriptor(self, ctx:QrogueDungeonParser.Collectible_descriptorContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#collectible_descriptor.
    def exitCollectible_descriptor(self, ctx:QrogueDungeonParser.Collectible_descriptorContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#energy_descriptor.
    def enterEnergy_descriptor(self, ctx:QrogueDungeonParser.Energy_descriptorContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#energy_descriptor.
    def exitEnergy_descriptor(self, ctx:QrogueDungeonParser.Energy_descriptorContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#enemy_descriptor.
    def enterEnemy_descriptor(self, ctx:QrogueDungeonParser.Enemy_descriptorContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#enemy_descriptor.
    def exitEnemy_descriptor(self, ctx:QrogueDungeonParser.Enemy_descriptorContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#boss_descriptor.
    def enterBoss_descriptor(self, ctx:QrogueDungeonParser.Boss_descriptorContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#boss_descriptor.
    def exitBoss_descriptor(self, ctx:QrogueDungeonParser.Boss_descriptorContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#riddle_descriptor.
    def enterRiddle_descriptor(self, ctx:QrogueDungeonParser.Riddle_descriptorContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#riddle_descriptor.
    def exitRiddle_descriptor(self, ctx:QrogueDungeonParser.Riddle_descriptorContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#challenge_descriptor.
    def enterChallenge_descriptor(self, ctx:QrogueDungeonParser.Challenge_descriptorContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#challenge_descriptor.
    def exitChallenge_descriptor(self, ctx:QrogueDungeonParser.Challenge_descriptorContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#puzzle_parameter.
    def enterPuzzle_parameter(self, ctx:QrogueDungeonParser.Puzzle_parameterContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#puzzle_parameter.
    def exitPuzzle_parameter(self, ctx:QrogueDungeonParser.Puzzle_parameterContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#input_stv.
    def enterInput_stv(self, ctx:QrogueDungeonParser.Input_stvContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#input_stv.
    def exitInput_stv(self, ctx:QrogueDungeonParser.Input_stvContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#boss_puzzle.
    def enterBoss_puzzle(self, ctx:QrogueDungeonParser.Boss_puzzleContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#boss_puzzle.
    def exitBoss_puzzle(self, ctx:QrogueDungeonParser.Boss_puzzleContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#circuit_stv.
    def enterCircuit_stv(self, ctx:QrogueDungeonParser.Circuit_stvContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#circuit_stv.
    def exitCircuit_stv(self, ctx:QrogueDungeonParser.Circuit_stvContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#circuit_gate_single.
    def enterCircuit_gate_single(self, ctx:QrogueDungeonParser.Circuit_gate_singleContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#circuit_gate_single.
    def exitCircuit_gate_single(self, ctx:QrogueDungeonParser.Circuit_gate_singleContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#circuit_gate_multi.
    def enterCircuit_gate_multi(self, ctx:QrogueDungeonParser.Circuit_gate_multiContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#circuit_gate_multi.
    def exitCircuit_gate_multi(self, ctx:QrogueDungeonParser.Circuit_gate_multiContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#draw_strategy.
    def enterDraw_strategy(self, ctx:QrogueDungeonParser.Draw_strategyContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#draw_strategy.
    def exitDraw_strategy(self, ctx:QrogueDungeonParser.Draw_strategyContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#stv_pools.
    def enterStv_pools(self, ctx:QrogueDungeonParser.Stv_poolsContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#stv_pools.
    def exitStv_pools(self, ctx:QrogueDungeonParser.Stv_poolsContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#default_stv_pool.
    def enterDefault_stv_pool(self, ctx:QrogueDungeonParser.Default_stv_poolContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#default_stv_pool.
    def exitDefault_stv_pool(self, ctx:QrogueDungeonParser.Default_stv_poolContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#stv_pool.
    def enterStv_pool(self, ctx:QrogueDungeonParser.Stv_poolContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#stv_pool.
    def exitStv_pool(self, ctx:QrogueDungeonParser.Stv_poolContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#stvs.
    def enterStvs(self, ctx:QrogueDungeonParser.StvsContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#stvs.
    def exitStvs(self, ctx:QrogueDungeonParser.StvsContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#stv_ref.
    def enterStv_ref(self, ctx:QrogueDungeonParser.Stv_refContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#stv_ref.
    def exitStv_ref(self, ctx:QrogueDungeonParser.Stv_refContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#stv.
    def enterStv(self, ctx:QrogueDungeonParser.StvContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#stv.
    def exitStv(self, ctx:QrogueDungeonParser.StvContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#reward_pools.
    def enterReward_pools(self, ctx:QrogueDungeonParser.Reward_poolsContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#reward_pools.
    def exitReward_pools(self, ctx:QrogueDungeonParser.Reward_poolsContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#default_reward_pool.
    def enterDefault_reward_pool(self, ctx:QrogueDungeonParser.Default_reward_poolContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#default_reward_pool.
    def exitDefault_reward_pool(self, ctx:QrogueDungeonParser.Default_reward_poolContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#reward_pool.
    def enterReward_pool(self, ctx:QrogueDungeonParser.Reward_poolContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#reward_pool.
    def exitReward_pool(self, ctx:QrogueDungeonParser.Reward_poolContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#collectibles.
    def enterCollectibles(self, ctx:QrogueDungeonParser.CollectiblesContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#collectibles.
    def exitCollectibles(self, ctx:QrogueDungeonParser.CollectiblesContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#collectible.
    def enterCollectible(self, ctx:QrogueDungeonParser.CollectibleContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#collectible.
    def exitCollectible(self, ctx:QrogueDungeonParser.CollectibleContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#integer.
    def enterInteger(self, ctx:QrogueDungeonParser.IntegerContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#integer.
    def exitInteger(self, ctx:QrogueDungeonParser.IntegerContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#complex_number.
    def enterComplex_number(self, ctx:QrogueDungeonParser.Complex_numberContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#complex_number.
    def exitComplex_number(self, ctx:QrogueDungeonParser.Complex_numberContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#layout.
    def enterLayout(self, ctx:QrogueDungeonParser.LayoutContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#layout.
    def exitLayout(self, ctx:QrogueDungeonParser.LayoutContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#l_room_row.
    def enterL_room_row(self, ctx:QrogueDungeonParser.L_room_rowContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#l_room_row.
    def exitL_room_row(self, ctx:QrogueDungeonParser.L_room_rowContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#l_hallway_row.
    def enterL_hallway_row(self, ctx:QrogueDungeonParser.L_hallway_rowContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#l_hallway_row.
    def exitL_hallway_row(self, ctx:QrogueDungeonParser.L_hallway_rowContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#rooms.
    def enterRooms(self, ctx:QrogueDungeonParser.RoomsContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#rooms.
    def exitRooms(self, ctx:QrogueDungeonParser.RoomsContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#room.
    def enterRoom(self, ctx:QrogueDungeonParser.RoomContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#room.
    def exitRoom(self, ctx:QrogueDungeonParser.RoomContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#r_attributes.
    def enterR_attributes(self, ctx:QrogueDungeonParser.R_attributesContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#r_attributes.
    def exitR_attributes(self, ctx:QrogueDungeonParser.R_attributesContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#r_visibility.
    def enterR_visibility(self, ctx:QrogueDungeonParser.R_visibilityContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#r_visibility.
    def exitR_visibility(self, ctx:QrogueDungeonParser.R_visibilityContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#r_type.
    def enterR_type(self, ctx:QrogueDungeonParser.R_typeContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#r_type.
    def exitR_type(self, ctx:QrogueDungeonParser.R_typeContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#hallways.
    def enterHallways(self, ctx:QrogueDungeonParser.HallwaysContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#hallways.
    def exitHallways(self, ctx:QrogueDungeonParser.HallwaysContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#hallway.
    def enterHallway(self, ctx:QrogueDungeonParser.HallwayContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#hallway.
    def exitHallway(self, ctx:QrogueDungeonParser.HallwayContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#h_attributes.
    def enterH_attributes(self, ctx:QrogueDungeonParser.H_attributesContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#h_attributes.
    def exitH_attributes(self, ctx:QrogueDungeonParser.H_attributesContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#messages.
    def enterMessages(self, ctx:QrogueDungeonParser.MessagesContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#messages.
    def exitMessages(self, ctx:QrogueDungeonParser.MessagesContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#message.
    def enterMessage(self, ctx:QrogueDungeonParser.MessageContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#message.
    def exitMessage(self, ctx:QrogueDungeonParser.MessageContext):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#message_body.
    def enterMessage_body(self, ctx:QrogueDungeonParser.Message_bodyContext):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#message_body.
    def exitMessage_body(self, ctx:QrogueDungeonParser.Message_bodyContext):
        pass



del QrogueDungeonParser


# exporting
from .door_tiles import DoorOpenState, DoorOneWayState, DoorEntanglementState, Door, HallwayEntrance
from .puzzle_tiles import Enemy, Boss
from .tiles import Invalid, Debug, Void, Floor, Wall, Obstacle, FogOfWar, Decoration, ControllableTile
from .tiles import TileCode, Tile
from .walk_trigger_tiles import Trigger, Teleport, Tunnel, Message, Challenger, Riddler, Collectible, Energy
from .walk_trigger_tiles import WalkTriggerTile

# importing
# +game.target_factory
# +logic
# +logic.actors
# +logic.actors.puzzles
# +logic.collectibles
# +world.navigation
# +util



# Generated from D:/Workspaces/pycharm-workspace/Qrogue/qrogue/management/save_grammar/SaveData.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .SaveDataParser import SaveDataParser
else:
    from SaveDataParser import SaveDataParser

# This class defines a complete generic visitor for a parse tree produced by SaveDataParser.

class SaveDataVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by SaveDataParser#start.
    def visitStart(self, ctx:SaveDataParser.StartContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SaveDataParser#date_time.
    def visitDate_time(self, ctx:SaveDataParser.Date_timeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SaveDataParser#duration.
    def visitDuration(self, ctx:SaveDataParser.DurationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SaveDataParser#inventory.
    def visitInventory(self, ctx:SaveDataParser.InventoryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SaveDataParser#gates.
    def visitGates(self, ctx:SaveDataParser.GatesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SaveDataParser#gate.
    def visitGate(self, ctx:SaveDataParser.GateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SaveDataParser#combined_gate.
    def visitCombined_gate(self, ctx:SaveDataParser.Combined_gateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SaveDataParser#background_gate.
    def visitBackground_gate(self, ctx:SaveDataParser.Background_gateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SaveDataParser#levels.
    def visitLevels(self, ctx:SaveDataParser.LevelsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SaveDataParser#level.
    def visitLevel(self, ctx:SaveDataParser.LevelContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SaveDataParser#score.
    def visitScore(self, ctx:SaveDataParser.ScoreContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SaveDataParser#value.
    def visitValue(self, ctx:SaveDataParser.ValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SaveDataParser#unlocks.
    def visitUnlocks(self, ctx:SaveDataParser.UnlocksContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SaveDataParser#unlock.
    def visitUnlock(self, ctx:SaveDataParser.UnlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SaveDataParser#achievements.
    def visitAchievements(self, ctx:SaveDataParser.AchievementsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SaveDataParser#achievement.
    def visitAchievement(self, ctx:SaveDataParser.AchievementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SaveDataParser#integer.
    def visitInteger(self, ctx:SaveDataParser.IntegerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SaveDataParser#complex_number.
    def visitComplex_number(self, ctx:SaveDataParser.Complex_numberContext):
        return self.visitChildren(ctx)



del SaveDataParser


from typing import List, Optional

from qiskit import Aer, execute, transpile, QuantumRegister, QuantumCircuit as QiskitCircuit
from qiskit.circuit import Gate

from qrogue.util import Config

if Config.using_new_qiskit():
    from qiskit.providers.basicaer import StatevectorSimulatorPy as StatevectorSimulator
else:
    from qiskit.providers.aer import StatevectorSimulator


class QuantumCircuit:
    @staticmethod
    def from_register(num_qubits: int) -> "QuantumCircuit":
        return QuantumCircuit(QiskitCircuit(QuantumRegister(num_qubits)))

    @staticmethod
    def from_bit_num(num_qubits: int, num_cbits: int = 0) -> "QuantumCircuit":
        return QuantumCircuit(QiskitCircuit(num_qubits, num_cbits))

    def __init__(self, circuit: QiskitCircuit):
        self.__circuit = circuit

    @property
    def circuit(self) -> QiskitCircuit:
        return self.__circuit

    def append(self, gate: Gate, qargs: List[int], cargs: List[int]):
        self.__circuit.append(gate, qargs, cargs)

    def to_gate(self, label: Optional[str] = None) -> Gate:
        return self.__circuit.to_gate(label=label)

    def copy(self, name: Optional[str] = None) -> "QuantumCircuit":
        return QuantumCircuit(self.__circuit.copy(name))

    def __str__(self) -> str:
        return str(self.__circuit)


class QuantumSimulator:
    def __init__(self):
        self.__simulator = StatevectorSimulator()

    def run(self, circuit: QuantumCircuit, do_transpile: bool = False) -> List[complex]:
        circuit: QiskitCircuit = circuit.circuit  # unwrap circuit

        if do_transpile:
            circuit = transpile(circuit, self.__simulator)

        # optionally use:   ddsim.JKQProvider().get_backend('statevector_simulator')?
        if Config.using_new_qiskit():
            # this signature doesn't take "shots=1"
            job = self.__simulator.run(circuit)
        else:
            # We only do 1 shot since we don't need any measurement but the StateVector
            job = self.__simulator.run(circuit, shots=1)

        return job.result().get_statevector()  # todo maybe pass circuit as experiment?


class UnitarySimulator:
    def __init__(self):
        self.__backend = Aer.get_backend('unitary_simulator')

    def execute(self, circuit: QuantumCircuit, decimals: int) -> List[List[complex]]:
        circuit: QiskitCircuit = circuit.circuit  # unwrap circuit

        job = execute(circuit, self.__backend)
        result = job.result()
        return result.get_unitary(circuit, decimals).data
